-------------------------------------------------------------------------------
--
-- SRAM controller
--
-- Aisha Ahmad
-- Akshay Deoras
-- Girish Gupta
-- Lab 6
--
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;

entity opb_bram is

  generic (
    C_OPB_AWIDTH : integer                   := 32;
    C_OPB_DWIDTH : integer                   := 32;
    --C_BASEADDR   : std_logic_vector(0 to 31) := X"0180_0000";
    --C_HIGHADDR   : std_logic_vector(0 to 31) := X"0180_3FFF");
    C_BASEADDR : std_logic_vector(0 to 31) := X"0086_0000";
    C_HIGHADDR : std_logic_vector(0 to 31) := X"0087_FFFF");

  port (
    OPB_Clk     : in  std_logic;
    OPB_Rst     : in  std_logic;
    OPB_ABus    : in  std_logic_vector(0 to C_OPB_AWIDTH-1);
    OPB_BE      : in  std_logic_vector(0 to C_OPB_DWIDTH/8-1);
    OPB_DBus    : in  std_logic_vector(0 to C_OPB_DWIDTH-1);
    OPB_RNW     : in  std_logic;
    OPB_select  : in  std_logic;
    OPB_seqAddr : in  std_logic;        -- Sequential Address
    Sln_DBus    : out std_logic_vector(0 to C_OPB_DWIDTH-1);
    Sln_errAck  : out std_logic;        -- (unused)
    Sln_retry   : out std_logic;        -- (unused)
    Sln_toutSup : out std_logic;        -- Timeout suppress
    Sln_xferAck : out std_logic;       -- Transfer acknowledge

    PB_D        : inout std_logic_vector(0 to 15);
    PB_A        : out std_logic_vector(0 to 17);
    PB_WE       : out std_logic;
    PB_OE       : out std_logic;
    PB_LB       : out std_logic;
    PB_UB       : out std_logic;
    PB_CE       : out std_logic
    );

end opb_bram;

architecture Behavioral of opb_bram is

  constant RAM_AWIDTH : integer := 18;  -- Number of address lines on the RAM
  constant RAM_DWIDTH : integer := 16;  -- Number of data lines on the RAM

  --component memoryctrl
    --port (
      --rst : in std_logic;
      --clk : in std_logic;
      --cs : in std_logic;
      --select0 : in std_logic;
      --rnw : in std_logic;
      --ce0 : out std_logic;
      --ce1 : out std_logic;
      --rres : out std_logic);
  --end component;

  --Modified code
  component OBUF_F_24
    port (
      O : out std_ulogic;
      I : in  std_ulogic);
  end component;

  component IOBUF_F_24
    port (
      O  : out   std_ulogic;
      IO : inout std_ulogic;
      I  : in    std_ulogic;
      T  : in    std_ulogic);
  end component;
  --End of modified code

  signal RNW : std_logic;
  signal RAM_DI, RAM_DO : std_logic_vector(0 to RAM_DWIDTH-1);
  signal ABus : std_logic_vector(0 to RAM_AWIDTH-1);
  signal chip_select : std_logic;
  signal output_enable : std_logic;
  signal WE, RST, CE, OE : std_logic;
  signal LB, UB : std_logic;
  signal tristate : std_logic;
  signal onecycle : std_logic;
  signal rce0, rce1, rreset : std_logic;

  -- Critical: Sln_xferAck is generated directly from state bit 0!
  constant STATE_BITS : integer := 3;
  constant Idle     : std_logic_vector(0 to STATE_BITS-1) := "000";
  constant Selected : std_logic_vector(0 to STATE_BITS-1) := "001";
  constant Read     : std_logic_vector(0 to STATE_BITS-1) := "011";
  constant Xfer     : std_logic_vector(0 to STATE_BITS-1) := "111";

  signal present_state, next_state : std_logic_vector(0 to STATE_BITS-1);

begin

  --modified port map
  addressbus: for i in 0 to 17 generate
    addresspad : OBUF_F_24 port map(
      O => PB_A(i),
      I => ABus(i)
    );
  end generate;

  databus: for j in 0 to 15 generate
    datapad : IOBUF_F_24 port map (
    O  => RAM_DO(j),
    IO => PB_D(j),
    I  => RAM_DI(j),
    T  => tristate);
  end generate;

  write_enable: OBUF_F_24 port map(
    O => PB_WE,
    I => WE
  );

  chip_enable: OBUF_F_24 port map(
    O => PB_CE,
    I => CE
  );

  o_enable: OBUF_F_24 port map(
    O => PB_OE,
    I => OE
  );

  lb_enable: OBUF_F_24 port map(
    O => PB_LB,
    I => LB
  );

  ub_enable: OBUF_F_24 port map(
    O => PB_UB,
    I => UB
  );

  -- Memory control/arbitration state machine

  --memoryctrl1 : memoryctrl port map (
    --rst => OPB_Rst,
    --clk => OPB_Clk,
    --cs => chip_select,
    --select0 => OPB_select,
    --rnw => RNW,
    --ce0 => rce0,
    --ce1 => rce1,
    --rres => rreset);

   --end of modified port map

  register_opb_inputs: process (OPB_Clk, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      RAM_DI <= (others => '0');							--data
      ABus <= (others => '0');
      RNW <= '0';
    elsif OPB_Clk'event and OPB_Clk = '1' then
      RAM_DI <= OPB_DBus(0 to RAM_DWIDTH-1);						--data
      ABus <= OPB_ABus(C_OPB_AWIDTH-3-(RAM_AWIDTH-1) to C_OPB_AWIDTH-3);
      RNW <= OPB_RNW;
    end if;
  end process register_opb_inputs;

  register_opb_outputs: process (OPB_Clk, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      Sln_DBus(0 to RAM_DWIDTH-1) <= (others => '0');					--data
    elsif OPB_Clk'event and OPB_Clk = '1' then
     if output_enable = '1' then
        Sln_DBus(0 to RAM_DWIDTH-1) <= RAM_DO;						--data
     else
       Sln_DBus(0 to RAM_DWIDTH-1) <= (others => '0');					--data
     end if;
    end if;
  end process register_opb_outputs;

    -- Write the low two bytes if rce0 or rce1 is enabled

  --register_opb_outputs1: process (OPB_Clk, OPB_Rst)
  --begin
    --if OPB_Rst = '1' then
      --Sln_DBus(0 to 15) <= X"0000";
    --elsif OPB_Clk'event and OPB_Clk = '1' then
      --if rreset = '1' then
        --Sln_DBus(0 to 15) <= X"0000";
      --elsif (rce1 or rce0) = '1' then
        --Sln_DBus(0 to 15) <= RAM_DO(0 to 15);
      --end if;
    --end if;
  --end process;

  -- Write the high two bytes if rce0 is enabled

  --register_opb_outputs2: process (OPB_Clk, OPB_Rst)
 -- begin
    --if OPB_Rst = '1' then
      --Sln_DBus(16 to 31) <= X"0000";
    --elsif OPB_Clk'event and OPB_Clk = '1' then
      --if rreset = '1' then
        --Sln_DBus(16 to 31) <= X"0000";
      --elsif rce0 = '1' then
       -- -Sln_DBus(16 to 31) <= RAM_DO(0 to 15);
      --end if;
    --end if;
  --end process;

  -- Unused outputs
  Sln_errAck  <= '0';
  Sln_retry   <= '0';
  Sln_toutSup <= '0';
  Sln_DBus(RAM_DWIDTH to C_OPB_DWIDTH-1) <= (others => '0');				--data

  --chip_select <=
    --'1' when OPB_select = '1' and
      --  OPB_ABus(0 to C_OPB_AWIDTH-3-RAM_AWIDTH) =
        --C_BASEADDR(0 to C_OPB_AWIDTH-3-RAM_AWIDTH) else
    --'0';

  chip_select <= OPB_select when OPB_ABus(31 downto 20) = X"008" else '0';

  -- Sequential part of the FSM
  fsm_seq : process(OPB_Clk, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      present_state <= Idle;
    elsif OPB_Clk'event and OPB_Clk = '1' then
      present_state <= next_state;
    end if;
  end process fsm_seq;

  -- Combinational part of the FSM
  fsm_comb : process(OPB_Rst, present_state, chip_select, OPB_Select, RNW)
  begin
    RST <= '1';                         -- Default values
    WE <= '1';                          -- modified
    output_enable <= '0';
    tristate <= '1';                    -- modified
    CE <= '0';
    OE <= '0';
    LB <= '0';
    UB <= '0';
    if OPB_RST = '1' then
      next_state <= Idle;
    else
      case present_state is
        when Idle =>
          if chip_select = '1' then
            next_state <= Selected;
          else
            next_state <= Idle;
          end if;

        when Selected =>
          if OPB_Select = '1' then
            if RNW = '1' then
              RST <= '0';
              output_enable <= '1';     -- modified
              tristate <= '1';
              next_state <= Read;
            else
              WE <= '0';                -- modified
              tristate <= '0';
              next_state <= Xfer;
            end if;
          else
            next_state <= Idle;
          end if;

        when Read =>
          if OPB_Select = '1' then
            -- tristate <= '1';         -- modified
            output_enable <= '1';
            next_state <= Xfer;
          else
            next_state <= Idle;
          end if;

        -- State encoding is critical here: xfer must only be true here
        when Xfer =>
          next_state <= Idle;

        when others =>
          next_state <= Idle;
      end case;
    end if;
  end process fsm_comb;

  Sln_xferAck <= present_state(0);

end Behavioral;

-- Local Variables:
-- compile-command: "ghdl -a opb_bram.vhd"
-- End:
